'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

class BehaviorData {
    constructor(combine, isStatic, id) {
        this.combine = combine;
        this.id = id;
        this.static = isStatic;
        this.empty = combine(none);
    }
    static get(behavior) {
        let value = behavior._data;
        if (!value)
            throw new RangeError("Not a behavior");
        return value;
    }
}
/// All extensions are associated with an extension group. This is
/// used to distinguish extensions meant for different types of hosts
/// (such as the editor view and state).
class ExtensionGroup {
    /// Create a new group. Client code probably doesn't need to do
    /// this. `getStore` retrieves the id-to-value map from a context
    /// object.
    constructor(getStore) {
        this.getStore = getStore;
        this.nextStorageID = 0;
        /// Mark an extension with a precedence below the default
        /// precedence, which will cause default-precedence extensions to
        /// override it even if they are specified later in the extension
        /// ordering.
        this.fallback = setPrec(-1 /* Fallback */);
        /// Mark an extension with normal precedence.
        this.normal = setPrec(0 /* Default */);
        /// Mark an extension with a precedence above the default precedence.
        this.extend = setPrec(1 /* Extend */);
        /// Mark an extension with a precedence above the default and
        /// `extend` precedences.
        this.override = setPrec(2 /* Override */);
    }
    behavior(options = {}) {
        let behavior = (value) => new ExtensionValue(0 /* Behavior */, behavior, { static: value }, this);
        behavior._data = new BehaviorData(options.combine || (array => array), !!options.static, this.storageID());
        return behavior;
    }
    /// Create an extension that adds a dynamically computed value for a
    /// given behavior. Dynamic behavior should usually just read and
    /// possibly transform a field from the context.
    dynamic(behavior, read) {
        if (BehaviorData.get(behavior).static)
            throw new Error("Can't create a dynamic source for a static behavior");
        return new ExtensionValue(0 /* Behavior */, behavior, { dynamic: read }, this);
    }
    /// Define a unique extension. When resolving extensions, all
    /// instances of a given unique extension are merged before their
    /// content extensions are retrieved. The `instantiate` function
    /// will be called with all the specs (configuration values) passed
    /// to the instances of the unique extension, and should resolve
    /// them to a more concrete extension value (or raise an error if
    /// they conflict).
    unique(instantiate, defaultSpec) {
        const type = new UniqueExtensionType(instantiate);
        return (spec = defaultSpec) => {
            if (spec === undefined)
                throw new RangeError("This extension has no default spec");
            return new ExtensionValue(2 /* Unique */, type, spec, this);
        };
    }
    /// Resolve an array of extensions by expanding all extensions until
    /// only behaviors are left, and then collecting the behaviors into
    /// arrays of values, preserving precedence ordering throughout.
    resolve(extensions) {
        return this.resolveInner(extensions);
    }
    /// @internal
    resolveInner(extensions, replace = none) {
        let pending = flatten(extensions, 0 /* Default */, replace);
        // This does a crude topological ordering to resolve behaviors
        // top-to-bottom in the dependency ordering. If there are no
        // cyclic dependencies, we can always find a behavior in the top
        // `pending` array that isn't a dependency of any unresolved
        // behavior, and thus find and order all its specs in order to
        // resolve them.
        for (let resolved = [];;) {
            let top = findTopUnique(pending, this);
            if (!top)
                break; // Only behaviors left
            // Prematurely evaluated a behavior type because of missing
            // sub-behavior information -- start over, in the assumption
            // that newly gathered information will make the next attempt
            // more successful.
            if (resolved.indexOf(top) > -1)
                return this.resolve(extensions);
            top.resolve(pending, replace);
            resolved.push(top);
        }
        // Collect the behavior values.
        let foreign = [];
        let readBehavior = Object.create(null);
        for (let ext of pending) {
            if (ext.type != this) {
                // Collect extensions of the wrong type into configuration.foreign
                foreign.push(ext);
                continue;
            }
            let behavior = BehaviorData.get(ext.id);
            if (Object.prototype.hasOwnProperty.call(readBehavior, behavior.id))
                continue; // Already collected
            let values = [];
            for (let e of pending)
                if (e.id == ext.id)
                    e.collect(values);
            let dynamic = [], parts = [];
            values.forEach(ext => {
                if (ext.value.dynamic) {
                    dynamic.push({ read: ext.value.dynamic, index: parts.length });
                    parts.push(null);
                }
                else {
                    parts.push(ext.value.static);
                }
            });
            if (dynamic.length == 0) {
                let value = behavior.combine(parts);
                readBehavior[behavior.id] = () => value;
            }
            else {
                let cached, cachedValue;
                readBehavior[behavior.id] = (context) => {
                    let values = this.getStore(context), found = values[behavior.id];
                    if (found !== undefined || Object.prototype.hasOwnProperty.call(values, behavior.id))
                        return found;
                    let array = parts.slice(), changed = false;
                    for (let { read, index } of dynamic) {
                        let newValue = array[index] = read(context);
                        if (!cached || cached[index] != newValue)
                            changed = true;
                    }
                    cached = array;
                    return values[behavior.id] = changed ? cachedValue = behavior.combine(array) : cachedValue;
                };
            }
        }
        return new Configuration(this, extensions, replace, readBehavior, foreign);
    }
    /// Define an extension name. Names can be used to tag extensions.
    /// This method returns a function that can be used to create named
    /// extensions, which can be used in a configuration as normal, but
    /// allow [replacement](#extension.Configuration.replaceExtensions)
    /// with another extension at some point in the future.
    defineName() {
        let name = (extension) => new ExtensionValue(3 /* Name */, name, extension, this);
        return name;
    }
    /// Allocate a unique storage number for use in field storage. Not
    /// something client code is likely to need.
    storageID() { return ++this.nextStorageID; }
}
function setPrec(prec) {
    return (extension) => extension instanceof ExtensionValue
        ? new ExtensionValue(extension.kind, extension.id, extension.value, extension.type, prec)
        : new ExtensionValue(1 /* Array */, null, extension, null, prec);
}
/// And extension is a value that describes a way in which something
/// is to be extended. It can be produced by instantiating a behavior,
/// calling unique extension function, or grouping extensions with
/// `Extension.all`.
class ExtensionValue {
    /// @internal
    constructor(
    /// @internal
    kind, 
    /// @internal
    id, 
    /// Holds the field for behaviors, the spec for unique extensions,
    /// and the array of extensions for multi extensions. @internal
    value, 
    /// @internal
    type, 
    /// @internal
    prec = -2 /* None */) {
        this.kind = kind;
        this.id = id;
        this.value = value;
        this.type = type;
        this.prec = prec;
    }
    // Insert this extension in an array of extensions so that it
    // appears after any already-present extensions with the same or
    // lower precedence, but before any extensions with higher
    // precedence.
    collect(array) {
        let i = 0;
        while (i < array.length && array[i].prec >= this.prec)
            i++;
        array.splice(i, 0, this);
    }
}
function flatten(extension, prec, replace, target = []) {
    if (Array.isArray(extension)) {
        for (let ext of extension)
            flatten(ext, prec, replace, target);
    }
    else {
        let value = extension;
        if (value.kind == 3 /* Name */) {
            let inner = value.value;
            for (let r of replace)
                if (r.id == value.id)
                    inner = r.value;
            flatten(inner, value.prec == -2 /* None */ ? prec : value.prec, replace, target);
        }
        else if (value.kind == 1 /* Array */) {
            for (let ext of value.value)
                flatten(ext, value.prec == -2 /* None */ ? prec : value.prec, replace, target);
        }
        else {
            target.push(value.prec != -2 /* None */ ? value : new ExtensionValue(value.kind, value.id, value.value, value.type, prec));
        }
    }
    return target;
}
class UniqueExtensionType {
    constructor(instantiate) {
        this.instantiate = instantiate;
        this.knownSubs = [];
    }
    hasSub(type) {
        for (let known of this.knownSubs)
            if (known == type || known.hasSub(type))
                return true;
        return false;
    }
    resolve(extensions, replace) {
        // Replace all instances of this type in extneions with the
        // sub-extensions that instantiating produces.
        let ours = [];
        for (let ext of extensions)
            if (ext.id == this)
                ext.collect(ours);
        let first = true;
        for (let i = 0; i < extensions.length; i++) {
            let ext = extensions[i];
            if (ext.id != this)
                continue;
            let sub = first ? this.subs(ours.map(s => s.value), ext.prec, replace) : none;
            extensions.splice(i, 1, ...sub);
            first = false;
            i += sub.length - 1;
        }
    }
    subs(specs, prec, replace) {
        let subs = flatten(this.instantiate(specs), prec, replace);
        for (let sub of subs)
            if (sub.kind == 2 /* Unique */ && this.knownSubs.indexOf(sub.id) == -1)
                this.knownSubs.push(sub.id);
        return subs;
    }
}
const none = [];
/// A configuration describes the fields and behaviors that exist in a
/// given set of extensions. It is created with
/// [`ExtensionGroup.resolve`](#extension.ExtensionGroup.resolve).
class Configuration {
    /// @internal
    constructor(type, extensions, replaced, readBehavior, 
    /// Any extensions that weren't an instance of the target
    /// extension group when resolving.
    foreign = []) {
        this.type = type;
        this.extensions = extensions;
        this.replaced = replaced;
        this.readBehavior = readBehavior;
        this.foreign = foreign;
    }
    /// Retrieve the value of a given behavior. When the behavior is
    /// [static](#extension.ExtensionGroup.behavior), the `context`
    /// argument can be omitted.
    getBehavior(behavior, context) {
        let data = BehaviorData.get(behavior);
        if (!context && !data.static)
            throw new RangeError("Need a context to retrieve non-static behavior");
        let f = this.readBehavior[data.id];
        return f ? f(context) : data.empty;
    }
    /// Replace one or more extensions with new ones, producing a new
    /// configuration.
    replaceExtensions(replace) {
        let repl = replace;
        for (let r of repl)
            if (r.kind != 3 /* Name */)
                throw new RangeError("Extension replacements must be named extension values");
        return this.type.resolveInner(this.extensions, this.replaced.filter(p => !repl.some(r => r.id == p.id)).concat(repl));
    }
}
// Find the extension type that must be resolved next, meaning it is
// not a (transitive) sub-extension of any other extensions that are
// still in extenders.
function findTopUnique(extensions, type) {
    let foundUnique = false;
    for (let ext of extensions)
        if (ext.kind == 2 /* Unique */ && ext.type == type) {
            foundUnique = true;
            if (!extensions.some(e => e.kind == 2 /* Unique */ && e.id.hasSub(ext.id)))
                return ext.id;
        }
    if (foundUnique)
        throw new RangeError("Sub-extension cycle in unique extensions");
    return null;
}
/// Utility function for combining behaviors to fill in a config
/// object from an array of provided configs. Will, by default, error
/// when a field gets two values that aren't ===-equal, but you can
/// provide combine functions per field to do something else.
function combineConfig(configs, defaults, // Should hold only the optional properties of Config, but I haven't managed to express that
combine = {}) {
    let result = {};
    for (let config of configs)
        for (let key of Object.keys(config)) {
            let value = config[key], current = result[key];
            if (current === undefined)
                result[key] = value;
            else if (current === value || value === undefined) ; // No conflict
            else if (Object.hasOwnProperty.call(combine, key))
                result[key] = combine[key](current, value);
            else
                throw new Error("Config merge conflict for field " + key);
        }
    for (let key in defaults)
        if (result[key] === undefined)
            result[key] = defaults[key];
    return result;
}
/// Defaults the fields in a configuration object to values given in
/// `defaults` if they are not already present.
function fillConfig(config, defaults) {
    let result = {};
    for (let key in config)
        result[key] = config[key];
    for (let key in defaults)
        if (result[key] === undefined)
            result[key] = defaults[key];
    return result;
}

exports.Configuration = Configuration;
exports.ExtensionGroup = ExtensionGroup;
exports.combineConfig = combineConfig;
exports.fillConfig = fillConfig;
