'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var state = require('../../state');
var view = require('../../view');
var extension = require('../../extension');
var gutter = require('../../gutter');

const foldAnnotation = state.Annotation.define();
function selectedLines(view) {
    let lines = [];
    for (let { head } of view.state.selection.ranges) {
        if (lines.some(l => l.from <= head && l.to >= head))
            continue;
        lines.push(view.lineAt(head));
    }
    return lines;
}
const foldCode = view => {
    if (!view.plugin(foldPlugin))
        return false;
    let fold = [];
    for (let line of selectedLines(view)) {
        let range = view.state.behavior(state.EditorState.foldable)
            .reduce((value, f) => value || f(view.state, line.from, line.to), null);
        if (range)
            fold.push(range);
    }
    if (!fold.length)
        return false;
    view.dispatch(view.state.t().annotate(foldAnnotation({ fold })));
    return true;
};
const unfoldCode = view => {
    let unfold = [], plugin = view.plugin(foldPlugin);
    if (!plugin)
        return false;
    for (let line of selectedLines(view)) {
        let folded = plugin.foldInside(line.from, line.to);
        if (folded)
            unfold.push(folded);
    }
    if (!unfold.length)
        return false;
    view.dispatch(view.state.t().annotate(foldAnnotation({ unfold })));
    return true;
};
const defaultConfig = {
    placeholderDOM: null,
    placeholderText: "…"
};
const foldConfigBehavior = view.EditorView.extend.behavior({
    combine(values) { return values.length ? values[0] : defaultConfig; }
});
window.beh = foldConfigBehavior;
const foldPlugin = view.ViewPlugin.create(view => new FoldPlugin(view)).decorations(p => p.decorations);
const codeFolding = view.EditorView.extend.unique((configs) => {
    return [
        foldPlugin.extension,
        foldConfigBehavior(extension.combineConfig(configs, defaultConfig)),
        view.EditorView.extend.fallback(view.EditorView.theme(defaultStyle))
    ];
}, {});
class FoldPlugin {
    constructor(view$1) {
        this.view = view$1;
        this.decorations = view.Decoration.none;
        let config = view$1.behavior(foldConfigBehavior);
        this.widgetConfig = { config, class: this.placeholderClass };
    }
    get placeholderClass() {
        return this.view.cssClass("foldPlaceholder");
    }
    update(update) {
        this.decorations = this.decorations.map(update.changes);
        for (let tr of update.transactions) {
            let ann = tr.annotation(foldAnnotation);
            if (ann)
                this.updateRanges(ann.fold || [], ann.unfold || []);
        }
        // Make sure widgets are redrawn with up-to-date classes
        if (update.themeChanged && this.placeholderClass != this.widgetConfig.class) {
            this.widgetConfig = { config: this.widgetConfig.config, class: this.placeholderClass };
            let deco = [], iter = this.decorations.iter(), next;
            while (next = iter.next())
                deco.push(view.Decoration.replace(next.from, next.to, { widget: new FoldWidget(this.widgetConfig) }));
            this.decorations = view.Decoration.set(deco);
        }
    }
    updateRanges(add, remove) {
        this.decorations = this.decorations.update(add.map(({ from, to }) => view.Decoration.replace(from, to, { widget: new FoldWidget(this.widgetConfig) })), remove.length ? (from, to) => !remove.some(r => r.from == from && r.to == to) : null, remove.reduce((m, r) => Math.min(m, r.from), 1e8), remove.reduce((m, r) => Math.max(m, r.to), 0));
    }
    foldInside(from, to) {
        let found = null;
        this.decorations.between(from, to, (from, to) => found = ({ from, to }));
        return found;
    }
}
class FoldWidget extends view.WidgetType {
    ignoreEvents() { return false; }
    toDOM(view) {
        let conf = this.value.config;
        if (conf.placeholderDOM)
            return conf.placeholderDOM();
        let element = document.createElement("span");
        element.textContent = conf.placeholderText;
        // FIXME should this have a role? does it make sense to allow focusing by keyboard?
        element.setAttribute("aria-label", view.phrase("folded code"));
        element.title = view.phrase("unfold");
        element.className = this.value.class;
        element.onclick = event => {
            let line = view.lineAt(view.posAtDOM(event.target));
            let folded = view.plugin(foldPlugin).foldInside(line.from, line.to);
            if (folded)
                view.dispatch(view.state.t().annotate(foldAnnotation({ unfold: [folded] })));
            event.preventDefault();
        };
        return element;
    }
}
const foldGutterDefaults = {
    openText: "▼",
    closedText: "▶"
};
class FoldMarker extends gutter.GutterMarker {
    constructor(config, open) {
        super();
        this.config = config;
        this.open = open;
    }
    eq(other) { return this.config == other.config && this.open == other.open; }
    toDOM(view) {
        let span = document.createElement("span");
        span.textContent = this.open ? this.config.openText : this.config.closedText;
        span.title = view.phrase(this.open ? "Fold line" : "Unfold line");
        return span;
    }
}
function foldGutter(config = {}) {
    let fullConfig = extension.fillConfig(config, foldGutterDefaults);
    return [
        new gutter.Gutter({
            style: "foldGutter",
            lineMarker(view, line) {
                // FIXME optimize this. At least don't run it for updates that
                // don't change anything relevant
                let plugin = view.plugin(foldPlugin);
                let folded = plugin.foldInside(line.from, line.to);
                if (folded)
                    return new FoldMarker(fullConfig, false);
                if (view.state.behavior(state.EditorState.foldable).some(f => f(view.state, line.from, line.to)))
                    return new FoldMarker(fullConfig, true);
                return null;
            },
            initialSpacer() {
                return new FoldMarker(fullConfig, false);
            },
            handleDOMEvents: {
                click: (view, line) => {
                    let plugin = view.plugin(foldPlugin);
                    let folded = plugin.foldInside(line.from, line.to);
                    if (folded) {
                        view.dispatch(view.state.t().annotate(foldAnnotation({ unfold: [folded] })));
                        return true;
                    }
                    let range = view.state.behavior(state.EditorState.foldable)
                        .reduce((value, f) => value || f(view.state, line.from, line.to), null);
                    if (range) {
                        view.dispatch(view.state.t().annotate(foldAnnotation({ fold: [range] })));
                        return true;
                    }
                    return false;
                }
            }
        }).extension,
        codeFolding()
    ];
}
const defaultStyle = {
    foldPlaceholder: {
        background: "#eee",
        border: "1px solid silver",
        color: "#888",
        borderRadius: ".2em",
        margin: "0 1px",
        padding: "0 1px",
        cursor: "pointer"
    },
    "gutterElement.foldGutter": {
        padding: "0 1px",
        cursor: "pointer"
    }
};

exports.codeFolding = codeFolding;
exports.foldCode = foldCode;
exports.foldGutter = foldGutter;
exports.unfoldCode = unfoldCode;
