'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var extension = require('../../extension');
var view = require('../../view');
var rangeset = require('../../rangeset');
var state = require('../../state');

/// A gutter marker represents a bit of information attached to a line
/// in a specific gutter. Your own custom markers have to extend this
/// class.
class GutterMarker extends rangeset.RangeValue {
    /// @internal
    compare(other) {
        return this == other || this.constructor == other.constructor && this.eq(other);
    }
    /// Map this marker through a position mapping.
    map(mapping, pos) {
        pos = mapping.mapPos(pos, -1, state.MapMode.TrackBefore);
        return pos < 0 ? null : new rangeset.Range(pos, pos, this);
    }
    /// Render the DOM node for this marker, if any.
    toDOM(view) { return null; }
    /// Create a range that places this marker at the given position.
    at(pos) { return new rangeset.Range(pos, pos, this); }
}
GutterMarker.prototype.elementClass = "";
const defaults = {
    style: "",
    renderEmptyElements: false,
    elementStyle: "",
    initialMarkers: () => rangeset.RangeSet.empty,
    updateMarkers: (markers) => markers,
    lineMarker: () => null,
    initialSpacer: null,
    updateSpacer: null,
    handleDOMEvents: {}
};
const gutterBehavior = view.EditorView.extend.behavior();
/// Defines an editor gutter.
class Gutter {
    constructor(config) {
        this.config = extension.fillConfig(config, defaults);
    }
    /// The extension that installs this gutter.
    get extension() {
        return [
            gutters(),
            gutterBehavior(this)
        ];
    }
}
/// The gutter-drawing plugin is automatically enabled when you add a
/// gutter, but you can use this function to explicitly configure it.
///
/// Unless `fixed` is explicitly set to `false`, the gutters are
/// fixed, meaning they don't scroll along with the content
/// horizontally.
const gutters = view.EditorView.extend.unique((config) => {
    let fixed = config.every(c => c.fixed !== false);
    return [
        view.ViewPlugin.create(view => new GutterView(view, { fixed }))
            .behavior(view.EditorView.scrollMargins, gutterView => gutterView.scrollMargins())
            .extension,
        view.EditorView.theme(baseTheme)
    ];
}, {});
class GutterView {
    constructor(view, config) {
        this.view = view;
        this.config = config;
        this.dom = document.createElement("div");
        this.dom.setAttribute("aria-hidden", "true");
        this.gutters = view.behavior(gutterBehavior).map(gutter => new SingleGutterView(view, gutter.config));
        for (let gutter of this.gutters)
            this.dom.appendChild(gutter.dom);
        if (config.fixed) {
            // FIXME IE11 fallback, which doesn't support position: sticky,
            // by using position: relative + event handlers that realign the
            // gutter (or just force fixed=false on IE11?)
            this.dom.style.position = "sticky";
        }
        view.scrollDOM.insertBefore(this.dom, view.contentDOM);
        this.updateTheme();
    }
    updateTheme() {
        this.dom.className = this.view.cssClass("gutters");
        for (let gutter of this.gutters)
            gutter.updateTheme();
    }
    update(update) {
        if (update.themeChanged)
            this.updateTheme();
        for (let gutter of this.gutters)
            gutter.update(update);
    }
    draw() {
        // FIXME would be nice to be able to recognize updates that didn't redraw
        let contexts = this.gutters.map(gutter => new UpdateContext(gutter, this.view.viewport));
        this.view.viewportLines(line => {
            let text;
            if (Array.isArray(line.type))
                text = line.type.find(b => b.type == view.BlockType.Text);
            else
                text = line.type == view.BlockType.Text ? line : undefined;
            if (!text)
                return;
            for (let cx of contexts)
                cx.line(this.view, text);
        }, 0);
        for (let cx of contexts)
            cx.finish(this.view);
        this.dom.style.minHeight = this.view.contentHeight + "px";
    }
    scrollMargins() {
        if (this.gutters.length == 0 || !this.config.fixed)
            return {};
        return getComputedStyle(this.view.scrollDOM).direction == "ltr" ? { left: this.dom.offsetWidth } : { right: this.dom.offsetWidth };
    }
}
class UpdateContext {
    constructor(gutter, viewport) {
        this.gutter = gutter;
        this.localMarkers = [];
        this.i = 0;
        this.height = 0;
        this.next = gutter.markers.iter(viewport.from, viewport.to).next;
        this.nextMarker = this.next();
    }
    line(view, line) {
        if (this.localMarkers.length)
            this.localMarkers = [];
        while (this.nextMarker && this.nextMarker.from <= line.from) {
            if (this.nextMarker.from == line.from)
                this.localMarkers.push(this.nextMarker.value);
            this.nextMarker = this.next();
        }
        let forLine = this.gutter.config.lineMarker(view, line, this.localMarkers);
        if (forLine)
            this.localMarkers.unshift(forLine);
        let gutter = this.gutter;
        if (this.localMarkers.length == 0 && !gutter.config.renderEmptyElements)
            return;
        let above = line.top - this.height;
        if (this.i == gutter.elements.length) {
            let newElt = new GutterElement(view, line.height, above, this.localMarkers, gutter.elementClass);
            gutter.elements.push(newElt);
            gutter.dom.appendChild(newElt.dom);
        }
        else {
            let markers = this.localMarkers, elt = gutter.elements[this.i];
            if (sameMarkers(markers, elt.markers)) {
                markers = elt.markers;
                this.localMarkers.length = 0;
            }
            elt.update(view, line.height, above, markers, gutter.elementClass);
        }
        this.height = line.bottom;
        this.i++;
    }
    finish(view) {
        let gutter = this.gutter;
        while (gutter.elements.length > this.i)
            gutter.dom.removeChild(gutter.elements.pop().dom);
    }
}
class SingleGutterView {
    constructor(view, config) {
        this.view = view;
        this.config = config;
        this.elements = [];
        this.spacer = null;
        this.dom = document.createElement("div");
        for (let prop in config.handleDOMEvents) {
            this.dom.addEventListener(prop, (event) => {
                let line = view.lineAtHeight(event.clientY);
                if (config.handleDOMEvents[prop](view, line, event))
                    event.preventDefault();
            });
        }
        this.markers = config.initialMarkers(view);
        if (config.initialSpacer) {
            this.spacer = new GutterElement(view, 0, 0, [config.initialSpacer(view)], this.elementClass);
            this.dom.appendChild(this.spacer.dom);
            this.spacer.dom.style.cssText += "visibility: hidden; pointer-events: none";
        }
        this.updateTheme();
    }
    updateTheme() {
        this.dom.className = this.view.cssClass("gutter" + (this.config.style ? "." + this.config.style : ""));
        this.elementClass = this.view.cssClass("gutterElement" + (this.config.style ? "." + this.config.style : ""));
        while (this.elements.length)
            this.dom.removeChild(this.elements.pop().dom);
    }
    update(update) {
        if (update.themeChanged)
            this.updateTheme();
        this.markers = this.config.updateMarkers(this.markers.map(update.changes), update);
        if (this.spacer && this.config.updateSpacer) {
            let updated = this.config.updateSpacer(this.spacer.markers[0], update);
            if (updated != this.spacer.markers[0])
                this.spacer.update(update.view, 0, 0, [updated], this.elementClass);
        }
    }
    destroy() {
        this.dom.remove();
    }
}
class GutterElement {
    constructor(view, height, above, markers, eltClass) {
        this.height = -1;
        this.above = 0;
        this.dom = document.createElement("div");
        this.update(view, height, above, markers, eltClass);
    }
    update(view, height, above, markers, cssClass) {
        if (this.height != height)
            this.dom.style.height = (this.height = height) + "px";
        if (this.above != above)
            this.dom.style.marginTop = (this.above = above) ? above + "px" : "";
        if (this.markers != markers) {
            this.markers = markers;
            for (let ch; ch = this.dom.lastChild;)
                ch.remove();
            let cls = cssClass;
            for (let m of markers) {
                let dom = m.toDOM(view);
                if (dom)
                    this.dom.appendChild(dom);
                let c = m.elementClass;
                if (c)
                    cls += " " + c;
            }
            this.dom.className = cls;
        }
    }
}
function sameMarkers(a, b) {
    if (a.length != b.length)
        return false;
    for (let i = 0; i < a.length; i++)
        if (!a[i].compare(b[i]))
            return false;
    return true;
}
/// Used to insert markers into the line number gutter.
const lineNumberMarkers = state.Annotation.define();
/// Create a line number gutter extension. The order in which the
/// gutters appear is determined by their extension priority.
const lineNumbers = view.EditorView.extend.unique(configs => {
    let config = extension.combineConfig(configs, { formatNumber: String, handleDOMEvents: {} }, {
        handleDOMEvents(a, b) {
            let result = {};
            for (let event in a)
                result[event] = a[event];
            for (let event in b) {
                let exists = result[event], add = b[event];
                result[event] = exists ? (view, line, event) => exists(view, line, event) || add(view, line, event) : add;
            }
            return result;
        }
    });
    class NumberMarker extends GutterMarker {
        constructor(number) {
            super();
            this.number = number;
        }
        eq(other) { return this.number == other.number; }
        toDOM() {
            return document.createTextNode(config.formatNumber(this.number));
        }
    }
    // FIXME preserve markers across reconfigurations by somehow making
    // this gutter static
    return new Gutter({
        style: "lineNumber",
        updateMarkers(markers, update) {
            let ann = update.annotation(lineNumberMarkers);
            if (ann)
                markers = markers.update(ann.add || [], ann.filter || null);
            return markers;
        },
        lineMarker(view, line, others) {
            if (others.length)
                return null;
            // FIXME try to make the line number queries cheaper?
            return new NumberMarker(view.state.doc.lineAt(line.from).number);
        },
        initialSpacer(view) {
            return new NumberMarker(maxLineNumber(view.state.doc.lines));
        },
        updateSpacer(spacer, update) {
            let max = maxLineNumber(update.view.state.doc.lines);
            return max == spacer.number ? spacer : new NumberMarker(max);
        }
    }).extension;
}, {});
function maxLineNumber(lines) {
    let last = 9;
    while (last < lines)
        last = last * 10 + 9;
    return last;
}
const baseTheme = {
    gutters: {
        background: "#f5f5f5",
        borderRight: "1px solid silver",
        color: "#999",
        display: "flex",
        height: "100%",
        boxSizing: "border-box",
        left: 0
    },
    gutter: {
        display: "flex !important",
        flexDirection: "column",
        flexShrink: 0,
        boxSizing: "border-box",
        height: "100%",
        overflow: "hidden"
    },
    gutterElement: {
        boxSizing: "border-box"
    },
    "gutterElement.lineNumber": {
        padding: "0 3px 0 5px",
        minWidth: "20px",
        textAlign: "right",
        whiteSpace: "nowrap"
    }
};

exports.Gutter = Gutter;
exports.GutterMarker = GutterMarker;
exports.gutters = gutters;
exports.lineNumberMarkers = lineNumberMarkers;
exports.lineNumbers = lineNumbers;
