'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var view = require('../../view');
var extension = require('../../extension');
var text = require('../../text');
var styleMod = require('style-mod');

/// Returns an extension that installs highlighting of special
/// characters.
const specialChars = view.EditorView.extend.unique((configs) => {
    // FIXME make configurations compose properly
    let config = extension.combineConfig(configs, {
        render: null,
        specialChars: SPECIALS,
        addSpecialChars: null
    });
    let styles = document.body.style;
    config.replaceTabs = (styles.tabSize || styles.MozTabSize) == null;
    if (config.replaceTabs)
        config.specialChars = new RegExp("\t|" + config.specialChars.source, "gu");
    let plugin = view.ViewPlugin.create(view => new SpecialCharPlugin(view, config))
        .decorations(plugin => plugin.decorations);
    return config.replaceTabs ? [plugin.extension, view.EditorView.styleModule(style)] : plugin.extension;
}, {});
const JOIN_GAP = 10;
class SpecialCharPlugin {
    constructor(view$1, options) {
        this.view = view$1;
        this.options = options;
        this.from = 0;
        this.to = 0;
        this.decorations = view.Decoration.none;
        this.specials = options.specialChars;
        if (options.addSpecialChars)
            this.specials = new RegExp(this.specials.source + "|" + options.addSpecialChars.source, "gu");
        this.updateForViewport();
    }
    update(update) {
        if (update.changes.length) {
            this.decorations = this.decorations.map(update.changes);
            this.from = update.changes.mapPos(this.from, 1);
            this.to = update.changes.mapPos(this.to, -1);
            this.closeHoles(update.changes.changedRanges());
        }
        this.updateForViewport();
    }
    closeHoles(ranges) {
        let decorations = [], vp = this.view.viewport, replaced = [];
        for (let i = 0; i < ranges.length; i++) {
            let { fromB: from, toB: to } = ranges[i];
            // Must redraw all tabs further on the line
            if (this.options.replaceTabs)
                to = this.view.state.doc.lineAt(to).end;
            while (i < ranges.length - 1 && ranges[i + 1].fromB < to + JOIN_GAP)
                to = Math.max(to, ranges[++i].toB);
            // Clip to current viewport, to avoid doing work for invisible text
            from = Math.max(vp.from, from);
            to = Math.min(vp.to, to);
            if (from >= to)
                continue;
            this.getDecorationsFor(from, to, decorations);
            replaced.push(from, to);
        }
        if (decorations.length)
            this.decorations = this.decorations.update(decorations, pos => {
                for (let i = 0; i < replaced.length; i += 2)
                    if (pos >= replaced[i] && pos < replaced[i + 1])
                        return false;
                return true;
            }, replaced[0], replaced[replaced.length - 1]);
    }
    updateForViewport() {
        let vp = this.view.viewport;
        // Viewports match, don't do anything
        if (this.from == vp.from && this.to == vp.to)
            return;
        let decorations = [];
        if (this.from >= vp.to || this.to <= vp.from) {
            this.getDecorationsFor(vp.from, vp.to, decorations);
            this.decorations = view.Decoration.set(decorations);
        }
        else {
            if (vp.from < this.from)
                this.getDecorationsFor(vp.from, this.from, decorations);
            if (this.to < vp.to)
                this.getDecorationsFor(this.to, vp.to, decorations);
            this.decorations = this.decorations.update(decorations, (from, to) => from >= vp.from && to <= vp.to);
        }
        this.from = vp.from;
        this.to = vp.to;
    }
    getDecorationsFor(from, to, target) {
        let { doc } = this.view.state;
        for (let pos = from, cursor = doc.iterRange(from, to), m; !cursor.next().done;) {
            if (!cursor.lineBreak) {
                while (m = this.specials.exec(cursor.value)) {
                    let code = m[0].codePointAt ? m[0].codePointAt(0) : m[0].charCodeAt(0), widget;
                    if (code == null)
                        continue;
                    if (code == 9) {
                        let line = doc.lineAt(pos + m.index);
                        let size = this.view.state.tabSize, col = text.countColumn(doc.slice(line.start, pos + m.index), 0, size);
                        widget = new TabWidget((size - (col % size)) * this.view.defaultCharacterWidth);
                    }
                    else {
                        widget = new SpecialCharWidget(this.options, code);
                    }
                    target.push(view.Decoration.replace(pos + m.index, pos + m.index + m[0].length, { widget }));
                }
            }
            pos += cursor.value.length;
        }
    }
}
const SPECIALS = /[\u0000-\u0008\u000a-\u001f\u007f-\u009f\u00ad\u061c\u200b-\u200f\u2028\u2029\ufeff\ufff9-\ufffc]/gu;
const NAMES = {
    0: "null",
    7: "bell",
    8: "backspace",
    10: "newline",
    11: "vertical tab",
    13: "carriage return",
    27: "escape",
    8203: "zero width space",
    8204: "zero width non-joiner",
    8205: "zero width joiner",
    8206: "left-to-right mark",
    8207: "right-to-left mark",
    8232: "line separator",
    8233: "paragraph separator",
    65279: "zero width no-break space",
    65532: "object replacement"
};
// Assigns placeholder characters from the Control Pictures block to
// ASCII control characters
function placeHolder(code) {
    if (code >= 32)
        return null;
    if (code == 10)
        return "\u2424";
    return String.fromCharCode(9216 + code);
}
const DEFAULT_PLACEHOLDER = "\u2022";
class SpecialCharWidget extends view.WidgetType {
    constructor(options, code) {
        super(code);
        this.options = options;
    }
    toDOM() {
        let ph = placeHolder(this.value) || DEFAULT_PLACEHOLDER;
        let desc = "Control character " + (NAMES[this.value] || this.value);
        let custom = this.options.render && this.options.render(this.value, desc, ph);
        if (custom)
            return custom;
        let span = document.createElement("span");
        span.textContent = ph;
        span.title = desc;
        span.setAttribute("aria-label", desc);
        span.style.color = "red";
        return span;
    }
    ignoreEvent() { return false; }
}
class TabWidget extends view.WidgetType {
    toDOM() {
        let span = document.createElement("span");
        span.textContent = "\t";
        span.className = style.tab;
        span.style.width = this.value + "px";
        return span;
    }
    ignoreEvent() { return false; }
}
const style = new styleMod.StyleModule({
    tab: {
        display: "inline-block",
        overflow: "hidden",
        verticalAlign: "bottom"
    }
});

exports.specialChars = specialChars;
