'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var state = require('../../state');
var lezerTree = require('lezer-tree');
var highlight = require('../../highlight');

// Counts the column offset in a string, taking tabs into account.
// Used mostly to find indentation.
// FIXME more or less duplicated in indent/src/indent.ts
function countColumn(string, end, tabSize, startIndex, startValue) {
    if (end == null) {
        end = string.search(/[^\s\u00a0]/);
        if (end == -1)
            end = string.length;
    }
    for (let i = startIndex || 0, n = startValue || 0;;) {
        let nextTab = string.indexOf("\t", i);
        if (nextTab < 0 || nextTab >= end)
            return n + (end - i);
        n += nextTab - i;
        n += tabSize - (n % tabSize);
        i = nextTab + 1;
    }
}
// STRING STREAM
/// Encapsulates a single line of input. Given to stream syntax code,
/// which uses it to tokenize the content.
class StringStream {
    /// @internal
    constructor(
    /// The line.
    string, 
    /// Current tab size.
    tabSize) {
        this.string = string;
        this.tabSize = tabSize;
        /// The current position on the line.
        this.pos = 0;
        /// The start position of the current token.
        this.start = 0;
        this.lineStart = 0;
        this.lastColumnPos = 0;
        this.lastColumnValue = 0;
    }
    /// True if we are at the end of the line.
    eol() { return this.pos >= this.string.length; }
    /// True if we are at the start of the line.
    sol() { return this.pos == this.lineStart; }
    /// Get the next code unit after the current position, or undefined
    /// if we're at the end of the line.
    peek() { return this.string.charAt(this.pos) || undefined; }
    /// Read the next code unit and advance `this.pos`.
    next() {
        if (this.pos < this.string.length)
            return this.string.charAt(this.pos++);
    }
    /// Match the next character against the given string, regular
    /// expression, or predicate. Consume and return it if it matches.
    eat(match) {
        let ch = this.string.charAt(this.pos);
        let ok;
        if (typeof match == "string")
            ok = ch == match;
        else
            ok = ch && (match instanceof RegExp ? match.test(ch) : match(ch));
        if (ok) {
            ++this.pos;
            return ch;
        }
    }
    /// Continue matching characters that match the given string,
    /// regular expression, or predicate function. Return true if any
    /// characters were consumed.
    eatWhile(match) {
        let start = this.pos;
        while (this.eat(match)) { }
        return this.pos > start;
    }
    /// Consume whitespace ahead of `this.pos`. Return true if any was
    /// found.
    eatSpace() {
        let start = this.pos;
        while (/[\s\u00a0]/.test(this.string.charAt(this.pos)))
            ++this.pos;
        return this.pos > start;
    }
    /// Move to the end of the line.
    skipToEnd() { this.pos = this.string.length; }
    /// Move to directly before the given character, if found on the
    /// current line.
    skipTo(ch) {
        let found = this.string.indexOf(ch, this.pos);
        if (found > -1) {
            this.pos = found;
            return true;
        }
    }
    /// Move back `n` characters.
    backUp(n) { this.pos -= n; }
    /// Get the column position at `this.pos`.
    column() {
        if (this.lastColumnPos < this.start) {
            this.lastColumnValue = countColumn(this.string, this.start, this.tabSize, this.lastColumnPos, this.lastColumnValue);
            this.lastColumnPos = this.start;
        }
        return this.lastColumnValue - (this.lineStart ? countColumn(this.string, this.lineStart, this.tabSize) : 0);
    }
    /// Get the indentation column of the current line.
    indentation() {
        return countColumn(this.string, null, this.tabSize) -
            (this.lineStart ? countColumn(this.string, this.lineStart, this.tabSize) : 0);
    }
    /// Match the input against the given string or regular expression
    /// (which should start with a `^`). Return true or the regexp match
    /// if it matches.
    ///
    /// Unless `consume` is set to `false`, this will move `this.pos`
    /// past the matched text.
    ///
    /// When matching a string `caseInsensitive` can be set to true to
    /// make the match case-insensitive.
    match(pattern, consume, caseInsensitive) {
        if (typeof pattern == "string") {
            let cased = (str) => caseInsensitive ? str.toLowerCase() : str;
            let substr = this.string.substr(this.pos, pattern.length);
            if (cased(substr) == cased(pattern)) {
                if (consume !== false)
                    this.pos += pattern.length;
                return true;
            }
            else
                return null;
        }
        else {
            let match = this.string.slice(this.pos).match(pattern);
            if (match && match.index > 0)
                return null;
            if (match && consume !== false)
                this.pos += match[0].length;
            return match;
        }
    }
    /// Get the current token.
    current() { return this.string.slice(this.start, this.pos); }
    /// Hide the first `n` characters of the stream while running
    /// `inner`. This can be useful for nesting modes.
    hideFirstChars(n, inner) {
        this.lineStart += n;
        try {
            return inner();
        }
        finally {
            this.lineStart -= n;
        }
    }
}
class StringStreamCursor {
    constructor(text, offset, tabSize = 4) {
        this.offset = offset;
        this.tabSize = tabSize;
        this.iter = text.iterLines(offset);
        this.curLineEnd = this.offset - 1;
    }
    next() {
        let { value, done } = this.iter.next();
        if (done)
            throw new RangeError("Reached end of document");
        const res = new StringStream(value, this.tabSize);
        this.offset = this.curLineEnd + 1;
        this.curLineEnd += value.length + 1;
        return res;
    }
}

class StreamParserInstance {
    constructor(spec) {
        this.token = spec.token;
        this.blankLine = spec.blankLine || (() => { });
        this.startState = spec.startState || (() => true);
        this.copyState = spec.copyState || defaultCopyState;
        this.indent = spec.indent || (() => -1);
        this.docType = docID(spec.docProps || []);
    }
    readToken(state, stream, editorState) {
        stream.start = stream.pos;
        for (let i = 0; i < 10; i++) {
            let result = this.token(stream, state, editorState);
            if (stream.pos > stream.start)
                return result;
        }
        throw new Error("Stream parser failed to advance stream.");
    }
}
function defaultCopyState(state) {
    if (typeof state != "object")
        return state;
    let newState = {};
    for (let prop in state) {
        let val = state[prop];
        newState[prop] = (val instanceof Array ? val.slice() : val);
    }
    return newState;
}
class RequestInfo {
    constructor(upto) {
        this.upto = upto;
        this.promise = new Promise(r => this.resolve = r);
        this.promise.canceled = false;
    }
}
/// A syntax provider that uses a stream parser.
class StreamSyntax {
    /// Create a stream syntax.
    constructor(parser) {
        this.parser = new StreamParserInstance(parser);
        this.field = new state.StateField({
            init: state => new SyntaxState(lezerTree.Tree.empty, [this.parser.startState(state)], 1, 0, null),
            apply: (tr, value) => value.apply(tr)
        });
        this.extension = [
            state.EditorState.syntax(this),
            this.field.extension,
            state.EditorState.indentation((state, pos) => {
                return state.field(this.field).getIndent(this.parser, state, pos);
            })
        ];
    }
    tryGetTree(state, from, to) {
        let field = state.field(this.field);
        return field.updateTree(this.parser, state, to, false) ? field.tree : null;
    }
    getTree(state, from, to) {
        let field = state.field(this.field);
        let rest = field.updateTree(this.parser, state, to, true);
        return { tree: field.tree, rest: rest === true ? null : rest };
    }
    getPartialTree(state, from, to) {
        let field = state.field(this.field);
        field.updateTree(this.parser, state, to, false);
        return field.tree;
    }
    docTypeAt(state, pos) {
        return typeArray[this.parser.docType];
    }
}
const CACHE_STEP_SHIFT = 6, CACHE_STEP = 1 << CACHE_STEP_SHIFT;
const MAX_RECOMPUTE_DISTANCE = 20e3;
const WORK_SLICE = 100, WORK_PAUSE = 200;
class SyntaxState {
    constructor(tree, 
    // Slot 0 stores the start state (line 1), slot 1 the
    // state at the start of line 65, etc, so lineNo ==
    // (index * CACHE_STEP) + 1
    cache, frontierLine, frontierPos, frontierState) {
        this.tree = tree;
        this.cache = cache;
        this.frontierLine = frontierLine;
        this.frontierPos = frontierPos;
        this.frontierState = frontierState;
        this.requests = [];
        this.working = -1;
    }
    apply(tr) {
        if (!tr.docChanged)
            return this;
        let { start, number } = tr.doc.lineAt(tr.changes.changedRanges()[0].fromA);
        if (number >= this.frontierLine)
            return new SyntaxState(this.tree, this.cache.slice(), this.frontierLine, this.frontierPos, this.frontierState);
        else {
            return new SyntaxState(this.tree.cut(start), this.cache.slice(0, (number >> CACHE_STEP_SHIFT) + 1), number, start, null);
        }
    }
    maybeStoreState(parser, lineBefore, state) {
        if (lineBefore % CACHE_STEP == 0)
            this.cache[(lineBefore - 1) >> CACHE_STEP_SHIFT] = parser.copyState(state);
    }
    findState(parser, editorState, line) {
        let cacheIndex = Math.min(this.cache.length - 1, (line - 1) >> CACHE_STEP_SHIFT);
        let cachedLine = (cacheIndex << CACHE_STEP_SHIFT) + 1;
        let startPos = editorState.doc.line(cachedLine).start;
        if (line - cachedLine > CACHE_STEP && editorState.doc.line(line).start - startPos > MAX_RECOMPUTE_DISTANCE)
            return null;
        let state = parser.copyState(this.cache[cacheIndex]);
        let cursor = new StringStreamCursor(editorState.doc, startPos, editorState.tabSize);
        for (let l = cachedLine; l < line; l++) {
            let stream = cursor.next();
            if (stream.eol()) {
                parser.blankLine(state, editorState);
            }
            else {
                while (!stream.eol())
                    parser.readToken(state, stream, editorState);
            }
            this.maybeStoreState(parser, l, state);
        }
        return state;
    }
    advanceFrontier(parser, editorState, upto) {
        let state = this.frontierState || this.findState(parser, editorState, this.frontierLine);
        let sliceEnd = Date.now() + WORK_SLICE;
        let cursor = new StringStreamCursor(editorState.doc, this.frontierPos, editorState.tabSize);
        let buffer = [];
        let line = this.frontierLine, pos = this.frontierPos;
        while (pos < upto) {
            let stream = cursor.next(), offset = cursor.offset;
            if (stream.eol()) {
                parser.blankLine(state, editorState);
            }
            else {
                while (!stream.eol()) {
                    let type = parser.readToken(state, stream, editorState);
                    if (type)
                        buffer.push(tokenID(type), offset + stream.start, offset + stream.pos, 4);
                }
            }
            this.maybeStoreState(parser, line, state);
            line++;
            pos += stream.string.length + 1;
            if (Date.now() > sliceEnd)
                break;
        }
        let tree = lezerTree.Tree.build(buffer, nodeGroup, parser.docType).balance();
        this.tree = this.tree.append(tree).balance();
        this.frontierLine = line;
        this.frontierPos = pos;
        this.frontierState = state;
    }
    updateTree(parser, state, upto, rest) {
        upto = Math.min(upto + 100, state.doc.lineAt(upto).end);
        // FIXME make sure multiple calls in same frame don't keep doing work
        if (this.frontierPos >= upto)
            return true;
        if (this.working == -1)
            this.advanceFrontier(parser, state, upto);
        if (this.frontierPos >= upto)
            return true;
        if (!rest)
            return false;
        let req = this.requests.find(r => r.upto == upto && !r.promise.canceled);
        if (!req) {
            req = new RequestInfo(upto);
            this.requests.push(req);
        }
        this.scheduleWork(parser, state);
        return req.promise;
    }
    scheduleWork(parser, state) {
        if (this.working != -1)
            return;
        this.working = setTimeout(() => this.work(parser, state), WORK_PAUSE);
    }
    work(parser, state) {
        this.working = -1;
        let upto = this.requests.reduce((max, req) => req.promise.canceled ? max : Math.max(max, req.upto), 0);
        if (upto > this.frontierPos)
            this.advanceFrontier(parser, state, upto);
        this.requests = this.requests.filter(req => {
            if (req.upto > this.frontierPos && !req.promise.canceled)
                return true;
            if (!req.promise.canceled)
                req.resolve(this.tree);
            return false;
        });
        if (this.requests.length)
            this.scheduleWork(parser, state);
    }
    getIndent(parser, state, pos) {
        let line = state.doc.lineAt(pos);
        let parseState = this.findState(parser, state, line.number);
        if (parseState == null)
            return -1;
        let text = line.slice(pos - line.start, Math.min(line.end, pos + 100) - line.start);
        return parser.indent(parseState, /^\s*(.*)/.exec(text)[1], state);
    }
}
const tokenTable = Object.create(null);
const typeArray = [lezerTree.NodeType.none];
const nodeGroup = new lezerTree.NodeGroup(typeArray);
const warned = [];
function tokenID(tag) {
    let id = tokenTable[tag];
    if (id == null) {
        let tagID = 0;
        try {
            tagID = highlight.defaultTags.get(tag);
        }
        catch (e) {
            if (!(e instanceof RangeError))
                throw e;
            if (warned.indexOf(tag) < 0) {
                warned.push(tag);
                console.warn(`'${tag}' is not a valid style tag`);
            }
            return tokenID("");
        }
        id = tokenTable[tag] = typeArray.length;
        typeArray.push(new lezerTree.NodeType(tag ? tag.replace(/ /g, "_") : "_", highlight.defaultTags.prop.set({}, tagID), id));
    }
    return id;
}
function docID(props) {
    if (props.length == 0)
        return tokenID("");
    let obj = Object.create(null);
    for (let [prop, value] of props)
        prop.set(obj, value);
    let id = typeArray.length;
    typeArray.push(new lezerTree.NodeType("document", obj, id));
    return id;
}

exports.StreamSyntax = StreamSyntax;
exports.StringStream = StringStream;
