'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var lezerTree = require('lezer-tree');
var state = require('../../state');

/// A syntax tree node prop used to associate indentation strategies
/// with node types. Such a strategy is a function from an indentation
/// context to a number. That number may be -1, to indicate that no
/// definitive indentation can be determined, or a column number to
/// which the given line should be indented.
const indentNodeProp = new lezerTree.NodeProp();
function syntaxIndentation(syntax) {
    return state.EditorState.indentation((state, pos) => {
        let tree = syntax.getPartialTree(state, pos, pos);
        return computeIndentation(state, tree, pos);
    });
}
// Compute the indentation for a given position from the syntax tree.
function computeIndentation(state, ast, pos) {
    let tree = ast.resolve(pos);
    // Enter previous nodes that end in empty error terms, which means
    // they were broken off by error recovery, so that indentation
    // works even if the constructs haven't been finished.
    for (let scan = tree, scanPos = pos;;) {
        let last = scan.childBefore(scanPos);
        if (!last)
            break;
        if (last.type.prop(lezerTree.NodeProp.error) && last.start == last.end) {
            tree = scan;
            scanPos = last.start;
        }
        else {
            scan = last;
            scanPos = scan.end + 1;
        }
    }
    for (; tree; tree = tree.parent) {
        let strategy = indentStrategy(tree.type) || (tree.parent == null ? topIndent : null);
        if (strategy)
            return strategy(new IndentContext(state, pos, tree));
    }
    return -1;
}
function indentStrategy(type) {
    let strategy = type.prop(indentNodeProp);
    if (!strategy) {
        let delim = type.prop(lezerTree.NodeProp.delim);
        if (delim)
            return delimitedIndent({ closing: delim.split(" ")[1] });
    }
    return strategy;
}
function topIndent() { return 0; }
/// Objects of this type provide context information and helper
/// methods to indentation functions.
class IndentContext {
    /// @internal
    constructor(
    /// The editor state.
    state, 
    /// The position at which indentation is being computed.
    pos, 
    /// The syntax tree node for which the indentation strategy is
    /// registered.
    node) {
        this.state = state;
        this.pos = pos;
        this.node = node;
    }
    /// The indent unit (number of spaces per indentation level).
    get unit() { return this.state.indentUnit; }
    /// Get the text directly after `this.pos`, either the entire line
    /// or the next 50 characters, whichever is shorter.
    get textAfter() {
        return this.state.doc.slice(this.pos, Math.min(this.pos + 50, this.state.doc.lineAt(this.pos).end)).match(/^\s*(.*)/)[1];
    }
    /// find the column position (taking tabs into account) of the given
    /// position in the given string.
    countColumn(line, pos) {
        // FIXME use extending character information
        if (pos < 0)
            pos = line.length;
        let tab = this.state.tabSize;
        for (var i = 0, n = 0;;) {
            let nextTab = line.indexOf("\t", i);
            if (nextTab < 0 || nextTab >= pos)
                return n + (pos - i);
            n += nextTab - i;
            n += tab - (n % tab);
            i = nextTab + 1;
        }
    }
    /// Find the indentation column of the given document line.
    lineIndent(line) {
        let text = line.slice(0, Math.min(50, line.length, this.node.start > line.start ? this.node.start - line.start : 1e8));
        return this.countColumn(text, text.search(/\S/));
    }
    /// Get the indentation at the reference line for `this.tree`, which
    /// is the line on which it starts, unless there is a node that is
    /// _not_ a parent of this node covering the start of that line. If
    /// so, the line at the start of that node is tried, again skipping
    /// on if it is covered by another such node.
    get baseIndent() {
        let line = this.state.doc.lineAt(this.node.start);
        // Skip line starts that are covered by a sibling (or cousin, etc)
        for (;;) {
            let atBreak = this.node.resolve(line.start);
            while (atBreak.parent && atBreak.parent.start == atBreak.start)
                atBreak = atBreak.parent;
            if (isParent(atBreak, this.node))
                break;
            line = this.state.doc.lineAt(atBreak.start);
        }
        return this.lineIndent(line);
    }
    /// Find the column for the given position.
    column(pos) {
        let line = this.state.doc.lineAt(pos);
        return this.countColumn(line.slice(0, pos - line.start), pos - line.start);
    }
}
function isParent(parent, of) {
    for (let cur = of; cur; cur = cur.parent)
        if (parent == cur)
            return true;
    return false;
}
// Check whether a delimited node is aligned (meaning there are
// non-skipped nodes on the same line as the opening delimiter). And
// if so, return the opening token.
function bracketedAligned(context) {
    let tree = context.node;
    let openToken = tree.childAfter(tree.start);
    if (!openToken)
        return null;
    let openLine = context.state.doc.lineAt(openToken.start);
    for (let pos = openToken.end;;) {
        let next = tree.childAfter(pos);
        if (!next)
            return null;
        if (!next.type.prop(lezerTree.NodeProp.skipped))
            return next.start < openLine.end ? openToken : null;
        pos = next.end;
    }
}
/// An indentation strategy for delimited (usually bracketed) nodes.
/// Will, by default, indent one unit more than the parent's base
/// indent unless the line starts with a closing token. When `align`
/// is true and there are non-skipped nodes on the node's opening
/// line, the content of the node will be aligned with the end of the
/// opening node, like this:
///
///     foo(bar,
///         baz)
function delimitedIndent({ closing, align = true, units = 1 }) {
    return (context) => {
        let closed = context.textAfter.slice(0, closing.length) == closing;
        let aligned = align ? bracketedAligned(context) : null;
        if (aligned)
            return closed ? context.column(aligned.start) : context.column(aligned.end);
        return context.baseIndent + (closed ? 0 : context.unit * units);
    };
}
/// An indentation strategy that aligns a node content to its base
/// indentation.
const flatIndent = (context) => context.baseIndent;
/// Creates an indentation strategy that, by default, indents
/// continued lines one unit more than the node's base indentation.
/// You can provide `except` to prevent indentation of lines that
/// match a pattern (for example `/^else\b/` in `if`/`else`
/// constructs), and you can change the amount of units used with the
/// `units` option.
function continuedIndent({ except, units = 1 } = {}) {
    return (context) => {
        let matchExcept = except && except.test(context.textAfter);
        return context.baseIndent + (matchExcept ? 0 : units * context.unit);
    };
}

/// This node prop is used to associate folding information with node
/// types. Given a subtree, it should check whether that tree is
/// foldable and return the range that can be collapsed when it is.
const foldNodeProp = new lezerTree.NodeProp();
function syntaxFolding(syntax) {
    return state.EditorState.foldable((state, start, end) => {
        let tree = syntax.getPartialTree(state, start, Math.min(state.doc.length, end + 100));
        let inner = tree.resolve(end);
        let found = null;
        for (let cur = inner; cur; cur = cur.parent) {
            if (cur.start < start || cur.end <= end)
                continue;
            let prop = cur.type.prop(foldNodeProp);
            if (prop) {
                let value = prop(cur);
                if (value && value.to > end)
                    found = value;
            }
        }
        return found;
    });
}

/// A [syntax provider](#state.Syntax) based on a
/// [Lezer](https://lezer.codemirror.net) parser.
class LezerSyntax {
    /// Create a syntax instance for the given parser. You'll usually
    /// want to use the
    /// [`withProps`](https://lezer.codemirror.net/docs/ref/#lezer.Parser.withProps)
    /// method to register CodeMirror-specific syntax node props in the
    /// parser, before passing it to this constructor.
    constructor(parser) {
        this.parser = parser;
        this.field = new state.StateField({
            init() { return new SyntaxState(lezerTree.Tree.empty); },
            apply(tr, value) { return value.apply(tr); }
        });
        this.extension = [state.EditorState.syntax(this), this.field.extension, syntaxIndentation(this), syntaxFolding(this)];
    }
    tryGetTree(state, from, to) {
        let field = state.field(this.field);
        return field.updateTree(this.parser, state.doc, from, to, false) ? field.tree : null;
    }
    getTree(state, from, to) {
        let field = state.field(this.field);
        let rest = field.updateTree(this.parser, state.doc, from, to, true);
        return { tree: field.tree, rest: rest === true ? null : rest };
    }
    getPartialTree(state, from, to) {
        let field = state.field(this.field);
        field.updateTree(this.parser, state.doc, from, to, false);
        return field.tree;
    }
    docTypeAt(state, pos) {
        if (this.parser.hasNested) {
            let tree = this.getPartialTree(state, pos, pos);
            let target = tree.resolve(pos);
            while (target) {
                if (target.type.prop(lezerTree.NodeProp.top))
                    return target.type;
                target = target.parent;
            }
        }
        return this.parser.group.types[1];
    }
}
class DocStream {
    constructor(doc, length = doc.length) {
        this.doc = doc;
        this.length = length;
        this.cursorPos = 0;
        this.string = "";
        this.cursor = doc.iter();
    }
    get(pos) {
        if (pos >= this.length)
            return -1;
        let stringStart = this.cursorPos - this.string.length;
        if (pos < stringStart || pos >= this.cursorPos) {
            if (pos < this.cursorPos) { // Reset the cursor if we have to go back
                this.cursor = this.doc.iter();
                this.cursorPos = 0;
            }
            this.string = this.cursor.next(pos - this.cursorPos).value;
            this.cursorPos = pos + this.string.length;
            stringStart = this.cursorPos - this.string.length;
        }
        return this.string.charCodeAt(pos - stringStart);
    }
    read(from, to) {
        let stringStart = this.cursorPos - this.string.length;
        if (from < stringStart || to >= this.cursorPos)
            return this.doc.slice(from, to);
        else
            return this.string.slice(from - stringStart, to - stringStart);
    }
    clip(at) {
        return new DocStream(this.doc, at);
    }
}
class RequestInfo {
    constructor(upto) {
        this.upto = upto;
        this.promise = new Promise(r => this.resolve = r);
        this.promise.canceled = false;
    }
}
class SyntaxState {
    constructor(tree) {
        this.tree = tree;
        this.parsedTo = 0;
        this.parse = null;
        this.working = -1;
        this.requests = [];
    }
    apply(tr) {
        return tr.docChanged ? new SyntaxState(this.tree.applyChanges(tr.changes.changedRanges())) : this;
    }
    // FIXME implement clearing out parts of the tree when it is too big
    updateTree(parser, doc, from, to, rest) {
        if (to <= this.parsedTo)
            return true;
        if (!this.parse) {
            this.parse = parser.startParse(new DocStream(doc), { cache: this.tree });
            this.continueParse(to);
        }
        if (this.parsedTo >= to)
            return true;
        if (!rest)
            return false;
        this.scheduleWork();
        let req = this.requests.find(r => r.upto == to && !r.promise.canceled);
        if (!req)
            this.requests.push(req = new RequestInfo(to));
        return req.promise;
    }
    continueParse(to) {
        let endTime = Date.now() + 100 /* Slice */;
        for (let i = 0;; i++) {
            let done = this.parse.advance();
            if (done) {
                this.parsedTo = 1e9;
                this.parse = null;
                this.tree = done;
                return;
            }
            if (i == 1000) {
                i = 0;
                if (Date.now() > endTime)
                    break;
            }
        }
        this.parsedTo = this.parse.pos;
        // FIXME somehow avoid rebuilding all the nodes that are already
        // in this.tree when this happens repeatedly
        this.tree = this.parse.forceFinish();
        if (this.parsedTo >= to)
            this.parse = null;
    }
    scheduleWork() {
        if (this.working != -1)
            return;
        this.working = setTimeout(() => this.work(), 200 /* Pause */);
    }
    work() {
        this.working = -1;
        let to = this.requests.reduce((max, req) => req.promise.canceled ? max : Math.max(max, req.upto), 0);
        if (to > this.parsedTo)
            this.continueParse(to);
        this.requests = this.requests.filter(req => {
            if (!req.promise.canceled && req.upto > this.parsedTo)
                return true;
            if (!req.promise.canceled)
                req.resolve(this.tree);
            return false;
        });
        if (this.requests.length)
            this.scheduleWork();
    }
}

function mkMatchProp() { return new lezerTree.NodeProp({ deserialize(str) { return str.split(" "); } }); }
/// A node prop that encodes information about which other nodes match
/// this node as delimiters. Should hold a space-separated list of
/// node names of the closing nodes that match this node.
const openNodeProp = mkMatchProp();
/// Like `openNodeProp`, but for closing nodes. Should hold a
/// space-separated list of opening node names that match this closing
/// delimiter.
const closeNodeProp = mkMatchProp();

exports.IndentContext = IndentContext;
exports.LezerSyntax = LezerSyntax;
exports.closeNodeProp = closeNodeProp;
exports.continuedIndent = continuedIndent;
exports.delimitedIndent = delimitedIndent;
exports.flatIndent = flatIndent;
exports.foldNodeProp = foldNodeProp;
exports.indentNodeProp = indentNodeProp;
exports.openNodeProp = openNodeProp;
